
' Orb! v1.0 beta
' (C) 2007 i-TECH and Kristopher Windsor

#define float double
#define cfloat cdbl

#define game_level_max 10

Const screenx = 800, screeny = 600, game_fps = 40

Const true = -1, false = 0, pi = 3.141592653589793

Const alien_max = 512, bomb_max = 256, bonus_max = 64, player_frame_max = 80, missile_max = 256, spawner_max = 16, star_max = 256, ufo_max = 64, text_max = 32
Const highscore_max = 8

Const original_graphic_size = 16, alien_size = 32, bomb_size = 32, bonus_size = 32, mapblock_size = 48, player_size = 48, spawner_size = 48, ufo_size = 32
Const mapblock_height = 12, mapblock_length_max = 512, sound_max = 5

Const game_display_voffset = 24, game_gravity = .1, missile_speed = 4

Const datapath = "data\", highscore_file = datapath & "scores.dat", progress_file = datapath & "save.dat"

#include once "fmod.bi"
#include once "fbgfx.bi"
#include once "image.bas"

'check for box collision; used for player, alien, ufo; used in WITH section to work with multiple object types
#macro physics(object_size, bounce_factor)
  Dim As Integer x1, y1, x2, y2, cc
  Dim As float tx, ty, embeddepth
  
  'calculate map coords
  x1 = Int((.p.x - object_size / 2) / mapblock_size + 1)
  y1 = Int((.p.y - object_size / 2) / mapblock_size + 1)
  x2 = Int((.p.x + object_size / 2 - 1) / mapblock_size + 1)
  y2 = Int((.p.y + object_size / 2 - 1) / mapblock_size + 1)
  
  For x As Integer = x1 To x2
    For y As Integer = y1 To y2
      If mapblock(x, y) = mapblock_enum.ground Then
        tx = (x - .5) * mapblock_size - .p.x
        ty = (y - .5) * mapblock_size - .p.y
        
        If Abs(tx) < (mapblock_size + object_size) / 2 And Abs(ty) < (mapblock_size + object_size) / 2 Then
          If Abs(ty) > Abs(tx) Then
            'bounce vertically
            If ty > 0 Then
              'bounce up (going up after collision)
              embeddepth = (mapblock_size + object_size) / 2 - ty
              If embeddepth > 0 Then
                cc = true
                If y > 1 Then If mapblock(x, y - 1) = mapblock_enum.ground Then cc = false
                If cc Then
                  .p.y -= embeddepth
                  .v.y *= -bounce_factor
                  If Abs(.v.y) < .2 Then .v.y = 0
                End If
              End If
            Else
              'bounce down
              embeddepth = (mapblock_size + object_size) / 2 + ty
              If embeddepth > 0 Then
                cc = true
                If y < mapblock_height Then If mapblock(x, y + 1) = mapblock_enum.ground Then cc = false
                If cc Then
                  .p.y += embeddepth
                  .v.y *= -bounce_factor
                  If Abs(.v.y) < .2 Then .v.y = 0
                End If
              End If
            End If
          Else
            'bounce horizontally
            If tx > 0 Then
              'bounce left
              embeddepth = (mapblock_size + object_size) / 2 - tx
              If embeddepth > 0 Then
                cc = true
                If x > 1 Then If mapblock(x - 1, y) = mapblock_enum.ground Then cc = false
                If cc Then
                  .p.x -= embeddepth
                  .v.x *= -bounce_factor
                  If Abs(.v.x) < .2 Then .v.x = 0
                End If
              End If
            Else
              'bounce right
              embeddepth = (mapblock_size + object_size) / 2 + tx
              If embeddepth > 0 Then
                cc = true
                If x < mapblock_length Then If mapblock(x + 1, y) = mapblock_enum.ground Then cc = false
                If cc Then
                  .p.x += embeddepth
                  .v.x *= -bounce_factor
                  If Abs(.v.x) < .2 Then .v.x = 0
                End If
              End If
            End If
          End If
        End If
      End If
    Next y
  Next x
#endmacro

Enum alien_enum
  still    = 1
  vertical = 2
  Random   = 3
End Enum

Enum mapblock_enum
  Space   = 0
  ground  = 1
  updraft = 2
End Enum

Enum sound_enum
  enemy_die    = 1
  enemy_fire   = 2
  player_crash = 3
  player_fire  = 4
  player_die   = 5
End Enum

Type coord
  As float x, y
End Type

Type alien_type
  As coord p, v
  As alien_enum Type
  As Integer hit, lives
  As Integer killme
End Type

Type bomb_type
  As coord p
  As float v
  As coord explode
  As Integer killme
End Type

Type bonus_type
  As coord p
  As Integer killme
End Type

Type missile_type
  As coord p, v
  As Integer killme
End Type

Type player_type
  As coord p, v
  As Integer score, hit, lives
  As Integer missiles
  As Integer frame_jump
  As float angle
End Type

Type text_type
  As String t
  As coord p
  As Integer ttl
End Type

Type spawner_type
  As coord p
  As Integer aliens
  As Integer killme
End Type

Type ufo_type
  As coord p, v
  As Integer hit, lives
  As Integer killme
End Type

Type highscore_type
  As String player
  As Integer score
End Type

Type sound_type
  As fmusic_module Ptr s
  As Integer delay
End Type

Dim Shared As Integer progress_level, progress_score
Dim Shared As Integer alien_total, bomb_total, bonus_total, mapblock_length, missile_total, star_total, spawner_total, text_total, ufo_total
Dim Shared As Integer game_won, game_quit, game_lost, game_display_offset, game_frame_total

Dim Shared As String menu_main(0 To 5), menu_ingame(0 To 2), menu_aftergame(0 To 8), menu_highscore(0 To highscore_max + 2), menu_rules(0 To 12)
Dim Shared As String game_level_title

Dim Shared As fb.image Ptr original_graphic(1 To 128)
Dim Shared As fb.image Ptr alien_graphic(1 To 3, 0 To 1)
Dim Shared As fb.image Ptr bomb_graphic(0 To 2)
Dim Shared As fb.image Ptr bonus_graphic(0 To 1)
Dim Shared As fb.image Ptr mapblock_graphic(1 To 3)
Dim Shared As fb.image Ptr player_graphic(0 To player_frame_max - 1)
Dim Shared As fb.image Ptr spawner_graphic
Dim Shared As fb.image Ptr ufo_graphic

Dim Shared As alien_type alien(1 To alien_max)
Dim Shared As bomb_type bomb(1 To bomb_max)
Dim Shared As bonus_type bonus(1 To bonus_max)
Dim Shared As missile_type missile(1 To missile_max)
Dim Shared As player_type player
Dim Shared As spawner_type spawner(1 To spawner_max)
Dim Shared As coord star(1 To star_max)
Dim Shared As text_type text(1 To text_max)
Dim Shared As ufo_type ufo(1 To ufo_max)

Dim Shared As highscore_type highscore(1 To highscore_max)
Dim Shared As mapblock_enum mapblock(mapblock_length_max, mapblock_height)
Dim Shared As sound_type sound(1 To sound_max)

Declare Sub utility_print (Byref t As String, Byval px As Integer, Byval py As Integer)
Declare Sub bomb_add (Byval x As Integer, Byval y As Integer)
Declare Sub spawner_add (Byval x As Integer, Byval y As Integer)
Declare Sub text_add (Byref t As String, Byval x As Integer, Byval y As Integer)
Declare Sub ufo_add (Byval x As Integer, Byval y As Integer)
Declare Sub highscore_display
Declare Sub sound_add (Byval s As sound_enum, Byval x As Integer)
Declare Function game_menu(options() As String, Byval options_total As Integer, Byval disabled As Integer = false) As Integer

Sub utility_clear_buffer
  'clear key buffer
  If Inkey > "" Or Multikey(&H48) Or Multikey(&H4D) Or Multikey(&H4B) Or Multikey(&H39) Then
    Screenlock
    Cls
    Line (0, 50) - Step(screenx - 1, 0), &HFF888888
    Line (50, 0) - Step(0, screeny - 1), &HFF888888
    utility_print Chr(4, 5, 32) & "PLEASE WAIT..." & Chr(32, 4, 5), 100, 100
    Screenunlock
    Do
      While Inkey > "" Or Multikey(&H48) Or Multikey(&H4D) Or Multikey(&H4B) Or Multikey(&H39): Sleep 50, 1: Wend
      Sleep 500, 1
    Loop Until Inkey = ""
  End If
End Sub

Sub utility_colorize (Byval graphic As fb.image Ptr, Byval c As Integer)
  Dim As Integer srce_size_x, srce_size_y
  
  srce_size_x = graphic -> Width
  srce_size_y = graphic -> height
  
  For x As Integer = 0 To srce_size_x - 1
    For y As Integer = 0 To srce_size_y - 1
      If Point(x, y, graphic) <> &HFFFF00FF Then Pset graphic, (x, y), c
    Next y
  Next x
End Sub

Sub utility_print (Byref t As String, Byval px As Integer, Byval py As Integer)
  For a As Integer = 1 To Len(t)
    Put (px + (a - 1) * original_graphic_size, py), original_graphic(Asc(Mid(t, a, 1)) + 1), trans
  Next a
End Sub

Sub utility_put (Byval x As Integer, Byval y As Integer, Byval graphic As fb.image Ptr)
  Dim As Integer srce_size_x, srce_size_y
  
  srce_size_x = graphic -> Width
  srce_size_y = graphic -> height
  
  Put (x - srce_size_x / 2 - game_display_offset, y - srce_size_y / 2 + game_display_voffset), graphic, trans
End Sub

Sub utility_sleep
  Static As float frame_timer, sleep_time
  
  sleep_time = 1000 / game_fps - (Timer - frame_timer) * 1000
  If sleep_time > 0 Then Sleep sleep_time, 1
  frame_timer = Timer
End Sub

Sub alien_start
  For a As Integer = 0 To 1
    For b As Integer = 1 To 3
      alien_graphic(b, a) = imagecreate(original_graphic_size * 2, original_graphic_size)
    Next b
    
    Put alien_graphic(alien_enum.still, a), (0, 0), original_graphic(5 + a * 16), Pset
    Put alien_graphic(alien_enum.still, a), (original_graphic_size, 0), original_graphic(6 + a * 16), Pset
    utility_colorize alien_graphic(alien_enum.still, a), &HFFFF8800
    
    Put alien_graphic(alien_enum.vertical, a), (0, 0), original_graphic(3 + a * 16), Pset
    Put alien_graphic(alien_enum.vertical, a), (original_graphic_size, 0), original_graphic(4 + a * 16), Pset
    utility_colorize alien_graphic(alien_enum.vertical, a), &HFF8888FF
    
    Put alien_graphic(alien_enum.random, a), (original_graphic_size / 2, 0), original_graphic(2 + a * 16), Pset
    utility_colorize alien_graphic(alien_enum.random, a), &HFFFF44CC
  Next a
End Sub

Sub alien_reset
  alien_total = 0
End Sub

Sub alien_add (Byval tpe As Integer, Byval x As Integer, Byval y As Integer)
  If alien_total = alien_max Then Exit Sub
  
  alien_total += 1
  With alien(alien_total)
    With .p
      .x = x
      .y = y
    End With
    With .v
      Select Case tpe
      Case alien_enum.still
        .x = 0
        .y = 0
      Case alien_enum.vertical
        .x = 0
        .y = 3
      Case alien_enum.random
        .x = (Rnd - Rnd) * 6
        .y = (Rnd - Rnd) * 6
      End Select
    End With
    .hit = false
    .lives = tpe '1 to 3 lives
    .type = tpe
    .killme = false
  End With
End Sub

Sub alien_move
  Dim As Integer a
  
  For a = 1 To alien_total
    With alien(a)
      If .killme = false Then
        'check lives / hit / deth
        If .hit Then
          .hit = false
          .lives -= 1
          If .lives <= 0 Then .killme = true
          text_add Chr(9, 10), alien(a).p.x - alien_size / 2, alien(a).p.y - alien_size / 2
          sound_add sound_enum.enemy_die, .p.x
        End If
      End If
      
      If .killme = false Then
        'adjust position
        .p.x += .v.x
        .p.y += .v.y
        
        'drop bomb
        If Int(Rnd * game_fps * 10) = 0 Then
          bomb_add .p.x, .p.y
          sound_add sound_enum.enemy_fire, .p.x
        End If
        
        physics(alien_size, 1)
      End If
    End With
  Next a
  
  a = 1
  While a <= alien_total And alien_total > 0
    If alien(a).killme Then
      player.score += alien(a).type * 20 - 10
      Swap alien(a), alien(alien_total)
      alien_total -= 1
    Else
      a += 1
    End If
  Wend
End Sub

Sub alien_display
  Dim As Integer y
  
  For a As Integer = 1 To alien_total
    With alien(a)
      If .killme = false Then
        If .type = alien_enum.still Then y = Sin(game_frame_total / 10) * 8 Else y = 0
        utility_put .p.x, .p.y + y, alien_graphic(.type, Abs((game_frame_total Mod game_fps) < game_fps / 2))
      End If
    End With
  Next a
End Sub

Sub alien_finish
  For a As Integer = 1 To 3
    For b As Integer = 0 To 1
      imagedestroy(alien_graphic(a, b))
    Next b
  Next a
End Sub

Sub bomb_start
  For a As Integer = 0 To 2
    bomb_graphic(a) = imagecreate(original_graphic_size, original_graphic_size)
    Put bomb_graphic(a), (0, 0), original_graphic(a + 25), Pset
  Next a
End Sub

Sub bomb_reset
  bomb_total = 0
End Sub

Sub bomb_add (Byval x As Integer, Byval y As Integer)
  If bomb_total = bomb_max Then Exit Sub
  
  bomb_total += 1
  With bomb(bomb_total)
    With .p
      .x = x
      .y = y
    End With
    .v = 0
    With .explode
      .x = 0
      .y = 0
    End With
    .killme = false
  End With
End Sub

Sub bomb_move
  Dim As Integer a, x, y
  
  For a = 1 To bomb_total
    With bomb(a)
      If .killme = false Then
        'adjust position
        .v += .2
        .p.y += .v
        
        'collide with player
        If Abs(player.p.x - .p.x) < (player_size + bomb_size) / 2 Then
          If Abs(player.p.y - .p.y) < (player_size + bomb_size) / 2 Then
            player.hit = true
            .killme = true
            .explode.x = bomb(a).p.x - original_graphic_size / 2
            .explode.y = bomb(a).p.y
          End If
        End If
        
        If .killme = false Then
          'collide with ground
          x = Int(.p.x / mapblock_size + 1)
          y = Int(.p.y / mapblock_size + 1)
          If mapblock(x, y) = mapblock_enum.ground Then
            .killme = true
            .explode.x = bomb(a).p.x - original_graphic_size / 2
            .explode.y = mapblock_size * Int(bomb(a).p.y / mapblock_size) - original_graphic_size
          End If
        End If
      End If
    End With
  Next a
  
  a = 1
  While a <= bomb_total And bomb_total > 0
    If bomb(a).killme Then
      x = Iif(Rnd < .5, 97, 99)
      text_add Chr(x, x + 1), bomb(a).explode.x, bomb(a).explode.y
      Swap bomb(a), bomb(bomb_total)
      bomb_total -= 1
    Else
      a += 1
    End If
  Wend
End Sub

Sub bomb_display
  For a As Integer = 1 To bomb_total
    With bomb(a)
      If .killme = false Then
        utility_put .p.x, .p.y, bomb_graphic((4 * game_frame_total / game_fps) Mod 3)
      End If
    End With
  Next a
End Sub

Sub bomb_finish
  For a As Integer = 0 To 2
    imagedestroy(bomb_graphic(a))
  Next a
End Sub

Sub bonus_start
  For a As Integer = 0 To 1
    bonus_graphic(a) = imagecreate(original_graphic_size, original_graphic_size)
    Put bonus_graphic(a), (0, 0), original_graphic(15 + a * 16), Pset
    utility_colorize bonus_graphic(a), &HFFFFFF00
  Next a
End Sub

Sub bonus_reset
  bonus_total = 0
End Sub

Sub bonus_add (Byval x As Integer, Byval y As Integer)
  If bonus_total = bonus_max Then Exit Sub
  
  bonus_total += 1
  With bonus(bonus_total)
    With .p
      .x = x
      .y = y
    End With
     .killme = false
  End With
End Sub

Sub bonus_move
  Dim As Integer a
  
  For a = 1 To bonus_total
    With bonus(a)
      If Abs(player.p.x - .p.x) < (player_size + bonus_size) / 2 Then
        If Abs(player.p.y - .p.y) < (player_size + bonus_size) / 2 Then
          .killme = true
          player.score += 60
          player.lives += 2
          player.missiles += 100
          text_add Chr(22, 23), bonus(a).p.x - original_graphic_size / 2, bonus(a).p.y - original_graphic_size / 2
          sound_add sound_enum.player_crash, .p.x
        End If
      End If
    End With
  Next a
  
  a = 1
  While a <= bonus_total And bonus_total > 0
    If bonus(a).killme Then
      Swap bonus(a), bonus(bonus_total)
      bonus_total -= 1
    Else
      a += 1
    End If
  Wend
End Sub

Sub bonus_display
  For a As Integer = 1 To bonus_total
    With bonus(a)
      If .killme = false Then
        utility_put .p.x, .p.y, bonus_graphic(Abs((game_frame_total Mod game_fps) < game_fps / 2))
      End If
    End With
  Next a
End Sub

Sub bonus_finish
  For a As Integer = 0 To 1
    imagedestroy(bonus_graphic(a))
  Next a
End Sub

Sub mapblock_start
  Dim As Integer x, y, c, r
  
  For a As Integer = 1 To 3
    mapblock_graphic(a) = imagecreate(mapblock_size, mapblock_size)
  Next a
  
  'ground (to do)
  'mapblock_ground_graphic = imagecreate(mapblock_size, mapblock_size, &HFFF88FFF)
  Line mapblock_graphic(mapblock_enum.ground), (0, 0) - (mapblock_size - 1, mapblock_size - 1), &HFF88FF88, BF
  For a As Integer = 1 To 128
    x = Int(Rnd * mapblock_size)
    y = Int(Rnd * mapblock_size)
    c = Rgb(&H88 - Rnd * 32, &HFF - Rnd * 32, &H88 - Rnd * 32)
    r = Int(Rnd * 10) + 4
    For x0 As Integer = x - mapblock_size To x + mapblock_size Step mapblock_size
      For y0 As Integer = y - mapblock_size To y + mapblock_size Step mapblock_size
        Circle mapblock_graphic(mapblock_enum.ground), (x0, y0), r, c,,,, F
      Next y0
    Next x0
  Next a
  
  'updraft (to do)
  'mapblock_updraft_graphic = imagecreate(mapblock_size, mapblock_size, &HFFFFFFFF)
  For x = 0 To mapblock_size - 1 Step original_graphic_size
    For y = 0 To mapblock_size - 1 Step mapblock_size / 2 'original_graphic_size
      Put mapblock_graphic(mapblock_enum.updraft), (x, y), original_graphic(Iif(y = 0, 16, 32)), Pset
    Next y
  Next x
End Sub

Sub mapblock_reset
  Dim As String l
  
  Open datapath & "levels/" & progress_level & ".dat" For Input As #1
  Line Input #1, game_level_title
  
  For a As Integer = 1 To mapblock_height
    Line Input #1, l
    If a = 1 Then mapblock_length = Len(l)
    For b As Integer = 1 To mapblock_length
      mapblock(b, a) = mapblock_enum.space 'default
      
      Select Case Mid(l, b, 1)
      Case "0"
        mapblock(b, a) = mapblock_enum.ground
      Case "1"
        mapblock(b, a) = mapblock_enum.updraft
      Case "B"
        'add bonus
        bonus_add (b - .5) * mapblock_size, (a - .5) * mapblock_size
      Case "S"
        'add alien
        alien_add alien_enum.still, (b - .5) * mapblock_size, (a - .5) * mapblock_size
      Case "V"
        'add alien
        alien_add alien_enum.vertical, (b - .5) * mapblock_size, (a - .5) * mapblock_size
      Case "R"
        'add alien
        alien_add alien_enum.random, (b - .5) * mapblock_size, (a - .5) * mapblock_size
      Case "U"
        'add ufo
        ufo_add (b - .5) * mapblock_size, (a - .5) * mapblock_size
      Case "-"
        spawner_add (b - .5) * mapblock_size, (a - .5) * mapblock_size
      'Case Else
        'error
      End Select
    Next b
  Next a
  Close #1
End Sub

Sub mapblock_display
  Dim As Integer a
  
  For x As Integer = 1 To mapblock_length
    a = (x - 1) * mapblock_size - game_display_offset
    If a + mapblock_size >= 0 And a - screenx < 0 Then
      For y As Integer = 1 To mapblock_height
        Select Case mapblock(x, y)
        Case 0
        Case mapblock_enum.updraft
          Put (a, (y - 1) * mapblock_size + game_display_voffset), mapblock_graphic(mapblock(x, y)), (0, game_frame_total Mod mapblock_size) - (mapblock_size - 1, mapblock_size - 1), trans
          Put (a, y * mapblock_size + game_display_voffset - (game_frame_total Mod mapblock_size)), mapblock_graphic(mapblock(x, y)), (0, 0) - (mapblock_size - 1, (game_frame_total Mod mapblock_size) - 1), trans
        Case Else
          Put (a, (y - 1) * mapblock_size + game_display_voffset), mapblock_graphic(mapblock(x, y)), trans
        End Select
      Next y
    End If
  Next x
End Sub

Sub mapblock_finish
  For a As Integer = 1 To 3
    imagedestroy(mapblock_graphic(a))
  Next a
End Sub

Sub missile_reset
  missile_total = 0
End Sub

Sub missile_add
  If missile_total + 5 > missile_max Then Exit Sub
  
  For a As float = player.angle To player.angle + 2 * pi - .01 Step 2 * pi / 5
    missile_total += 1
    With missile(missile_total)
      With .p
        .x = player.p.x + Cos(a) * player_size / 2
        .y = player.p.y + Sin(a) * player_size / 2
      End With
      With .v
        .x = Cos(a) * missile_speed
        .y = Sin(a) * missile_speed
      End With
      .killme = false
    End With
  Next a
End Sub

Sub missile_move
  Dim As Integer a, x, y
  
  For a = 1 To missile_total
    With missile(a)
      If .killme = false Then
        'adjust position
        .p.x += .v.x
        .p.y += .v.y
        
        x = .p.x
        y = .p.y
        For b As Integer = 1 To alien_total
          If alien(b).p.x - alien_size / 2 < x Then
            If alien(b).p.x + alien_size / 2 > x Then
              If alien(b).p.y - alien_size / 2 < y Then
                If alien(b).p.y + alien_size / 2 > y Then
                  If .killme = false Then
                    alien(b).hit = true
                    .killme = true
                  End If
                End If
              End If
            End If
          End If
        Next b
        For b As Integer = 1 To ufo_total
          If ufo(b).p.x - ufo_size / 2 < x Then
            If ufo(b).p.x + ufo_size / 2 > x Then
              If ufo(b).p.y - ufo_size / 2 < y Then
                If ufo(b).p.y + ufo_size / 2 > y Then
                  If .killme = false Then
                    ufo(b).hit = true
                    .killme = true
                  End If
                End If
              End If
            End If
          End If
        Next b
        For b As Integer = 1 To spawner_total
          If spawner(b).p.x - spawner_size / 2 < x Then
            If spawner(b).p.x + spawner_size / 2 > x Then
              If spawner(b).p.y - spawner_size / 2 < y Then
                If spawner(b).p.y + spawner_size / 2 > y Then
                  If .killme = false Then
                    spawner(b).killme = true
                    .killme = true
                  End If
                End If
              End If
            End If
          End If
        Next b
        
        'see if missile hit ground
        x = Int(.p.x / mapblock_size + 1)
        y = Int(.p.y / mapblock_size + 1)
        If mapblock(x, y) = mapblock_enum.ground Then .killme = true
      End If
    End With
  Next a
  
  a = 1
  While a <= missile_total And missile_total > 0
    If missile(a).killme Then
      Swap missile(a), missile(missile_total)
      missile_total -= 1
    Else
      a += 1
    End If
  Wend
End Sub

Sub missile_display
  For a As Integer = 1 To missile_total
    With missile(a)
      If .killme = false Then
        Line (.p.x - game_display_offset, .p.y + game_display_voffset) - Step(-.v.x, -.v.y), &HFFFF0000
      End If
    End With
  Next a
End Sub

Sub player_start
  Dim As float q, r
  Dim As fb.image Ptr temp_player_graphic, temp_player_graphic2, temp_player_graphic3
  
  temp_player_graphic = imagecreate(original_graphic_size * 2, original_graphic_size)
  temp_player_graphic2 = imagecreate(original_graphic_size * 8, original_graphic_size * 4)
  temp_player_graphic3 = imagecreate(320, 320)
  
  Put temp_player_graphic, (0, 0), original_graphic(23), Pset
  Put temp_player_graphic, (original_graphic_size, 0), original_graphic(24), Pset
  multiput temp_player_graphic2, original_graphic_size * 4, original_graphic_size * 2, temp_player_graphic, 4, 4

  Screenlock
  For q As float = 0 To player_frame_max - 1
    Line temp_player_graphic3, (0, 0) - (319, 319), &HFFFF00FF, BF
    For r As float = pi * q / 40 To pi * 2 + pi * q / 40 Step pi * 2 / 5
      multiput temp_player_graphic3, 160 + Cos(r) * 116, 160 + Sin(r) * 116, temp_player_graphic2, 1.2, 1.2, r + pi / 2, true
    Next r
    
    player_graphic(q) = imagecreate(player_size, player_size)
    scaler player_graphic(q),,, temp_player_graphic3, player_size / 320
    
    If (Cint(q) Mod 8) = 0 Then
      Cls
      Put (q * (screenx + player_size) / (player_frame_max - 7) - 160, screeny / 2 - 160), temp_player_graphic3, trans
      utility_print "ORB", (screenx - original_graphic_size * 3) / 2, screeny * .4
      Screenunlock
      Sleep 10
      Screenlock
    End If
  Next q
  Screenunlock
  
  imagedestroy(temp_player_graphic)
  imagedestroy(temp_player_graphic2)
  imagedestroy(temp_player_graphic3)
End Sub

Sub player_reset
  With player
    .p.x = mapblock_size * 1.5
    .p.y = mapblock_size * 1.5
    .v.x = 0
    .v.y = 0
    .score = 0
    .hit = false
    .lives = 5
    .missiles = 100
    .frame_jump = 0
    .angle = 0
  End With
End Sub

Sub player_move
  Dim As Integer isinupdraft
  
  With player
    'check win / lose
    If .hit Then
      .hit = false
      .lives -= 1
      If .lives <= 0 Then game_lost = true
      sound_add sound_enum.player_die, .p.x
    End If
    If alien_total = 0 And ufo_total = 0 And spawner_total = 0 Then game_won = true
    
    'adjust position
    .p.x += .v.x
    .p.y += .v.y
    
    physics(player_size, .8)
    
    'fire missiles
    If (game_frame_total Mod 12) = 0 Then .missiles += 1
    If Multikey(&H39) And .missiles > 0 Then
      .missiles -= 1
      missile_add
      sound_add sound_enum.player_fire, .p.x
    End If
    
    'see if player is in updraft
    For x As Integer = x1 To x2
      For y As Integer = y1 To y2
        If mapblock(x, y) = mapblock_enum.updraft Then isinupdraft = true
      Next y
    Next x
    
    'adjust velocity (has to come after physics because physics has to run before the isinupdraft is calculated)
    If isinupdraft Then
      .v.y -= game_gravity
    Else
      .v.y += game_gravity
    End If
    .v.x += (Abs(Multikey(&H4D)) - Abs(Multikey(&H4B))) / 4
    If Multikey(&H48) Then
      If game_frame_total - .frame_jump > 80 Then
        .frame_jump = game_frame_total
        .v.y -= 4
      End If
    End If
    If Abs(.v.x) > 16 Then .v.x = 16 * Sgn(.v.x)
    If Abs(.v.y) > 16 Then .v.y = 16 * Sgn(.v.y)
    
    'adjust angle
    If Abs(.v.x) < .4 Then
      If .v.x = 0 Then .angle += .4  / 80 Else .angle += .4 * Sgn(.v.x) / 80
    Else
      .angle += .v.x / 80
    End If
    If .angle < 0 Then .angle += 2 * pi
  End With
End Sub

Sub player_display
  With player
    utility_put .p.x, .p.y, player_graphic(Int(80 * .angle / (2 * pi)) Mod player_frame_max)
  End With
End Sub

Sub player_finish
  For a As Integer = 0 To player_frame_max - 1
    imagedestroy(player_graphic(a))
  Next a
End Sub

Sub spawner_start
  spawner_graphic = imagecreate(spawner_size, spawner_size)
  For x As Integer = 0 To 2
    For y As Integer = 0 To 1
      Put spawner_graphic, (original_graphic_size * x - 2, original_graphic_size * (y + 1)), original_graphic(x + y * 16 + 12), Pset
    Next y
  Next x
  utility_colorize spawner_graphic, &HFF884400
End Sub

Sub spawner_reset
  spawner_total = 0
End Sub

Sub spawner_add (Byval x As Integer, Byval y As Integer)
  If spawner_total = spawner_max Then Exit Sub
  
  spawner_total += 1
  With spawner(spawner_total)
    With .p
      .x = x
      .y = y
    End With
    .aliens = 50
    .killme = false
  End With
End Sub

Sub spawner_move
  Dim As Integer a
  
  For a = 1 To spawner_total
    With spawner(a)
      If .killme = false Then
        'spawn random alien
        If .aliens > 0 Then
          If Int(Rnd * game_fps * 8) = 0 Then
            .aliens -= 1
            alien_add alien_enum.random, .p.x, .p.y
          End If
        End If
        
        'collide with player
        If Abs(.p.x - player.p.x) < (player_size + spawner_size) / 2 Then
          If Abs(.p.y - player.p.y) < (player_size + spawner_size) / 2 Then
            player.hit = true
            .killme = true
            sound_add sound_enum.player_crash, .p.x
          End If
        End If
      End If
    End With
  Next a
  
  a = 1
  While a <= spawner_total And spawner_total > 0
    If spawner(a).killme Then
      text_add Chr(11, 12, 13), spawner(a).p.x - spawner_size / 2, spawner(a).p.y - spawner_size / 6
      text_add Chr(27, 28, 29), spawner(a).p.x - spawner_size / 2, spawner(a).p.y + spawner_size / 6
      player.score += 25
      For b As Integer = 1 To 20
        alien_add alien_enum.random, spawner(a).p.x, spawner(a).p.y
      Next b
      Swap spawner(a), spawner(spawner_total)
      spawner_total -= 1
    Else
      a += 1
    End If
  Wend
End Sub

Sub spawner_display
  For a As Integer = 1 To spawner_total
    With spawner(a)
      If .killme = false Then
        utility_put .p.x, .p.y, spawner_graphic
      End If
    End With
  Next a
End Sub

Sub spawner_finish
  imagedestroy(spawner_graphic)
End Sub

Sub star_reset
  star_total = mapblock_length / 2
  If star_total > star_max Then star_total = star_max
  
  For a As Integer = 1 To star_total
    With star(a)
      .x = Rnd * (mapblock_size * mapblock_length / 2 - 8 + screenx / 2)
      .y = Rnd * (mapblock_size * 10 - 16) + mapblock_size
    End With
  Next a
End Sub

Sub star_display
  For a As Integer = 1 To star_total
    With star(a)
      Put (.x - game_display_offset / 2, .y + game_display_voffset), original_graphic(Asc("+") + 1), alpha, 80
    End With
  Next a
End Sub

Sub text_reset
  text_total = 0
End Sub

Sub text_add (Byref t As String, Byval x As Integer, Byval y As Integer)
  If text_total = text_max Then Exit Sub
  
  text_total += 1
  With text(text_total)
    .t = t
    With .p
      .x = x
      .y = y
    End With
    .ttl = game_fps / 2
  End With
End Sub

Sub text_move
  Dim As Integer a
  
  For a = 1 To text_total
    With text(a)
      If .ttl > 0 Then .ttl -= 1
    End With
  Next a
  
  a = 1
  While a <= text_total And text_total > 0
    If text(a).ttl = 0 Then
      Swap text(a), text(text_total)
      text_total -= 1
    Else
      a += 1
    End If
  Wend
End Sub

Sub text_display
  If (game_frame_total Mod (game_fps / 5)) < 4 Then Exit Sub
  
  For a As Integer = 1 To text_total
    With text(a)
      utility_print .t, .p.x - game_display_offset, .p.y + game_display_voffset
    End With
  Next a
End Sub

Sub ufo_start
  ufo_graphic = imagecreate(original_graphic_size * 2, original_graphic_size)
  Put ufo_graphic, (0, 0), original_graphic(7), Pset
  Put ufo_graphic, (original_graphic_size, 0), original_graphic(8), Pset
  utility_colorize ufo_graphic, &HFFFF0000
End Sub

Sub ufo_reset
  ufo_total = 0
End Sub

Sub ufo_add (Byval x As Integer, Byval y As Integer)
  If ufo_total = ufo_max Then Exit Sub
  
  ufo_total += 1
  With ufo(ufo_total)
    With .p
      .x = x
      .y = y
    End With
    With .v
        .x = 1
        .y = 0
    End With
    .hit = false
    .lives = 4
    .killme = false
  End With
End Sub

Sub ufo_move
  Dim As Integer a, isinupdraft
  
  For a = 1 To ufo_total
    With ufo(a)
      If .killme = false Then
        'check lives / hit / deth
        If .hit Then
          .hit = false
          .lives -= 1
          If .lives <= 0 Then .killme = true
          sound_add sound_enum.enemy_die, .p.x
        End If
      End If
      
      If .killme = false Then
        'adjust position
        .p.x += .v.x
        .p.y += .v.y
        
        physics(ufo_size, 1)
        
        'see if ufo is in updraft
        isinupdraft = false
        For x As Integer = x1 To x2
          For y As Integer = y1 To y2
            If mapblock(x, y) = mapblock_enum.updraft Then isinupdraft = true
          Next y
        Next x
    
        'adjust velocity
        If isinupdraft Then .v.y -= game_gravity Else .v.y += game_gravity
        
        'collide with player
        If Abs(.p.x - player.p.x) < (player_size + ufo_size) / 2 Then
          If Abs(.p.y - player.p.y) < (player_size + ufo_size) / 2 Then
            player.hit = true
            .killme = true
            sound_add sound_enum.enemy_die, .p.x
            sound_add sound_enum.player_crash, .p.x
          End If
        End If
      End If
    End With
  Next a
  
  a = 1
  While a <= ufo_total And ufo_total > 0
    If ufo(a).killme Then
      text_add "25", ufo(a).p.x - ufo_size / 2, ufo(a).p.y - ufo_size / 2
      player.score += 25
      Swap ufo(a), ufo(ufo_total)
      ufo_total -= 1
    Else
      a += 1
    End If
  Wend
End Sub

Sub ufo_display
  For a As Integer = 1 To ufo_total
    With ufo(a)
      If .killme = false Then
        utility_put .p.x, .p.y, ufo_graphic
      End If
    End With
  Next a
End Sub

Sub ufo_finish
  imagedestroy(ufo_graphic)
End Sub

'end of main game components

Sub highscore_start
  Open highscore_file For Append As #1: Close #1
  
  Open highscore_file For Input As #1
  For a As Integer = 1 To highscore_max
    With highscore(a)
      If Not Eof(1) Then Input #1, .player
      If Not Eof(1) Then Input #1, .score
    End With
  Next a
  Close #1
  
  menu_highscore(0) = "HIGH SCORES"
  menu_highscore(highscore_max + 2) = "DONE"
End Sub

Sub highscore_add (Byval score As Integer)
  Dim As Integer rank, fc
  Dim As String t, k
  
  'find rank
  For a As Integer = 1 To highscore_max
    With highscore(a)
      
      If .score < score Then rank = a: Exit For
    End With
  Next a
  
  If rank = 0 Then Exit Sub
  
  utility_clear_buffer
  
  'input name
  Do
    fc += 1
    
    If k = Chr(8) Then
      If Len(t) Then t = Left(t, Len(t) - 1)
    Else
      If Len(k) = 1 And Len(t) < 16 Then t += Ucase(k)
    End If
    
    Screenlock
    Cls
    Line (0, 50) - Step(screenx - 1, 0), &HFF888888
    Line (50, 0) - Step(0, screeny - 1), &HFF888888
    utility_print Chr(4, 5) & " HIGH SCORE! " & Chr(4, 5), 100, 100
    utility_print t & "_", 100, 140
    Put (screenx - 100 - 80, screeny - 100 - 80), player_graphic(fc Mod player_frame_max), trans
    Screenunlock
    
    Sleep 50
    k = Inkey
  Loop Until k = Chr(13)
  
  'bump lower scores
  For a As Integer = highscore_max - 1 To rank Step -1
    Swap highscore(a), highscore(a + 1)
  Next a
  
  'save new score
  With highscore(rank)
    .player = t
    .score = score
  End With
  
  'show scores
  highscore_display
End Sub

Sub highscore_display
  For a As Integer = 1 To highscore_max
    With highscore(a)
      menu_highscore(a) = Left(.player & Space(16), 16) & "  " & Right(String(7, "0") & .score, 8)
    End With
  Next a
  game_menu menu_highscore(), highscore_max + 2, true
End Sub

Sub highscore_finish
  Open highscore_file For Output As #1
  For a As Integer = 1 To highscore_max
    With highscore(a)
      Print #1, .player
      Print #1, .score
    End With
  Next a
  Close #1
End Sub

Sub progress_start
  Open progress_file For Append As #1: Close #1
  
  Open progress_file For Input As #1
  If Not Eof(1) Then Input #1, progress_level
  If Not Eof(1) Then Input #1, progress_score
  Close #1
End Sub

Sub progress_reset
  progress_level = 1
  progress_score = 0
End Sub

Sub progress_finish
  Open progress_file For Output As #1
  Print #1, progress_level
  Print #1, progress_score
  Close #1
End Sub

Sub sound_start
  If fsound_init(44100, 32, 0) = false Then Print "FMOD Error!": System
  
  For a As Integer = 1 To sound_max
    With sound(a)
      .s = fsound_sample_load(fsound_free, datapath & "sound/" & a & ".mp3", 0, 0, 0)
    End With
  Next a
End Sub

Sub sound_reset
  For a As Integer = 1 To sound_max
    With sound(a)
      .delay = 0
    End With
  Next a
End Sub

Sub sound_add (Byval s As sound_enum, Byval x As Integer)
  With sound(s)
    If .delay = 0 Then
      If x - game_display_offset + 12 > 0 And x - game_display_offset - 12 <= screenx Then
        fsound_playsound(fsound_free, .s)
        .delay = game_fps / 5
      End If
    End If
  End With
End Sub

Sub sound_move
  For a As Integer = 1 To sound_max
    With sound(a)
      If .delay > 0 Then .delay -= 1
    End With
  Next a
End Sub

Sub sound_finish
  fmusic_stopallsongs
  fsound_close
End Sub

Function game_menu(options() As String, Byval options_total As Integer, Byval disabled As Integer = false) As Integer
  Static fc As Integer
  
  Dim As Integer k, s
  Dim As String t
  
  utility_clear_buffer
  
  s = Iif(disabled, options_total, 1)
  Do
    fc += 1
    
    If disabled = false Then
      Select Case k
      Case 72, 75
        Do
          If s > 1 Then s -= 1
        Loop Until options(s) > ""
      Case 77, 80
        Do
          If s < options_total Then s += 1
        Loop Until options(s) > ""
      End Select
    End If
    
    Screenlock
    Cls
    Line (0, 50) - Step(screenx - 1, 0), &HFF888888
    Line (50, 0) - Step(0, screeny - 1), &HFF888888
    utility_print Chr(4, 5, 32) & options(0) & Chr(32, 4, 5), 100, 100
    For a As Integer = 1 To options_total
      t = options(a)
      If a = s Then t = Chr(2, 3, 32) & t Else t = "   " & t
      utility_print t, 100, 120 + a * 20
    Next a
    Put (screenx - 100 - 80, screeny - 100 - 80), player_graphic(fc Mod player_frame_max), trans
    Screenunlock
    
    Sleep 50
    k = Asc(Right(Inkey, 1))
  Loop Until k = 32 Or k = 13
  
  Return s
End Function

Sub game_start
  Dim As Integer a, b, c, d, e, sx, sy, fx, fy
  Dim As fb.image Ptr original_image
  
  Screenres screenx, screeny, 32
  Randomize Timer
  
  original_image = imagecreate(128, 64)
  Bload datapath & "graphics.bmp", original_image
  
  For a = 0 To 15
    For b = 0 To 7
      c = b * 16 + a + 1
      original_graphic(c) = imagecreate(original_graphic_size, original_graphic_size)
      sx = a * 8
      sy = b * 8
      For d = 0 To original_graphic_size - 1
        For e = 0 To original_graphic_size - 1
          fx = Int(d / 2) + sx
          fy = Int(e / 2) + sy
          Pset original_graphic(c), (d, e), Point(fx, fy, original_image)
        Next e
      Next d
    Next b
  Next a
  
  sound_start
  alien_start
  bomb_start
  bonus_start
  mapblock_start
  player_start
  spawner_start
  ufo_start
  highscore_start
  progress_start
  
  'menus
  menu_main(0) = "ORB"
  menu_main(1) = "CONTINUE GAME"
  menu_main(2) = "NEW GAME"
  menu_main(3) = "HIGH SCORES"
  menu_main(4) = "HOW TO PLAY"
  menu_main(5) = "QUIT"
  
  menu_ingame(0) = "ORB"
  menu_ingame(1) = "CONTINUE GAME"
  menu_ingame(2) = "QUIT"
  
  menu_rules(0) = "HOW TO PLAY"
  menu_rules(1) = "<>  MOVE THE ORB WITH THE ARROW KEYS"
  menu_rules(2) = "^   JUMP WITH THE UP ARROW KEY"
  menu_rules(3) = Chr(16) & "   FIRE WITH THE SPACE BAR"
  menu_rules(4) = Chr(24, 25, 26) & " SHOOT THE ALIENS AND UFOS!"
  menu_rules(5) = "     " & Chr(4, 5) & " 10 POINTS"
  menu_rules(6) = "     " & Chr(6, 7) & " 25 POINTS"
  menu_rules(7) = "     " & Chr(2, 3) & " 30 POINTS"
  menu_rules(8) = "     " & Chr(1) & "  50 POINTS"
  menu_rules(9) = Chr(14, 30) & "  BONUS 60 POINTS 100 MISSILES 2 LIVES"
  menu_rules(10) = Chr(11, 12, 13) & " DESTROY THE ALIEN SPAWNERS"
  menu_rules(11) = ""
  menu_rules(12) = "DONE"
  
  imagedestroy(original_image)
End Sub

Sub game_reset
  If progress_level = 0 Then progress_reset
  
  game_won = false
  game_quit = false
  game_lost = false
  
  game_frame_total = 0
  
  'some objects need to be reset before the mapblocks
  alien_reset
  bomb_reset
  bonus_reset
  spawner_reset
  ufo_reset
  mapblock_reset
  
  missile_reset
  player_reset
  star_reset
  text_reset
  sound_reset
End Sub

Sub game_move
  Dim As Integer x
  
  game_frame_total += 1
  
  If Inkey = Chr(27) Then
    Select Case game_menu(menu_ingame(), 2)
    Case 1
    Case 2
      game_quit = true
    End Select
  End If
  
  'set camera
  x = player.p.x - game_display_offset
  If x / screenx < .3 Then game_display_offset -= (screenx * .3 - x) / 8
  If x / screenx > .7 Then game_display_offset += (x - screenx * .7) / 8
  If game_display_offset < 0 Then game_display_offset = 0
  If game_display_offset + screenx > mapblock_length * mapblock_size Then game_display_offset = mapblock_length * mapblock_size - screenx
  
  bomb_move
  player_move
  missile_move
  spawner_move
  alien_move
  ufo_move
  bonus_move
  text_move
  sound_move
End Sub

Sub game_display
  Dim As String t
  
  Screenlock
  Cls
  star_display
  mapblock_display
  spawner_display
  player_display
  alien_display
  ufo_display
  bonus_display
  text_display
  bomb_display
  missile_display
  
  If (Int(game_frame_total / (game_fps * 4)) Mod 5) < 3 Then
    t = "ORB LEVEL " & progress_level
  Else
    t = "[" & game_level_title & "]"
  End If
  t += " SCORE: " & player.score
  utility_print t, 0, 4
  
  t = Chr(8, 32) & player.missiles & Chr(32, 22, 23, 32) & player.lives
  If alien_total > 0 Then t += Chr(32, 1, 32) & alien_total
  If ufo_total > 0 Then t += Chr(32, 6, 7, 32) & ufo_total
  utility_print t, screenx - 1 - Len(t) * original_graphic_size, 4
  Screenunlock
End Sub

Sub game_over
  Dim As Integer nc, score, timebonus
  
  If game_lost And game_won Then game_lost = false
  
  'show stats
  
  If game_quit Then
    score = player.score
    menu_aftergame(0) = "GAME TERMINATED!"
    menu_aftergame(1) = "LEVEL SCORE: " & score
    nc = 1
  End If
  
  If game_lost Then
    score = progress_score + player.score
    menu_aftergame(0) = "GAME OVER!"
    menu_aftergame(1) = "LEVEL SCORE: " & player.score
    menu_aftergame(2) = "TOTAL SCORE: " & score
    nc = 2
  End If
  
  If game_won Then
    timebonus = 4000 - game_frame_total
    If timebonus < 0 Then timebonus = 0
    score = progress_score + player.score + player.lives * 75 + Iif(progress_level = game_level_max, 5500, 500) + timebonus
    If progress_level = game_level_max Then
      menu_aftergame(0) = "GAME COMPLETE!"
    Else
      menu_aftergame(0) = "LEVEL COMPLETE!"
    End If
    menu_aftergame(1) = "SUB SCORE        " & player.score
    menu_aftergame(2) = "LIFE BONUS       " & (player.lives * 75)
    menu_aftergame(3) = "TIME BONUS       " & timebonus
    menu_aftergame(4) = "COMPLETION BONUS " & Iif(progress_level = game_level_max, 5500, 500)
    menu_aftergame(5) = "LEVEL SCORE      " & score - progress_score
    menu_aftergame(6) = "TOTAL SCORE      " & score
    nc = 6
  End If
  
  nc += 2
  menu_aftergame(nc - 1) = ""
  
  If score > highscore(highscore_max).score And (game_won = false Or progress_level = game_level_max) Then
    menu_aftergame(nc) = "ENTER HIGH SCORES"
  Else
    menu_aftergame(nc) = "DONE"
  End If
  game_menu(menu_aftergame(), nc, true)
  
  If game_quit Then
    'if the player quits, they lose the level, but their progress is kept (not sent back to level 1), so they are not illegible for a highscore
    'they can only get a highscore with the points from this level, not counting the combined scores of the previous levels
    highscore_add score
  End If
  
  If game_lost Then
    progress_reset
    highscore_add score
  End If
  
  If game_won Then
    progress_level += 1
    progress_score = score
    
    If progress_level > game_level_max Then
      highscore_add score
      progress_reset
    End If
  End If
End Sub

Sub game_play
  'called for every level
  
  Dim As float sleep_time, v
  
  game_reset
  
  'level intro
  game_display_offset = mapblock_length * mapblock_size - screenx
  If game_display_offset < 0 Then game_display_offset = 0
  While game_display_offset > 0
    v += .2
    game_display
    game_display_offset -= v
    utility_sleep
  Wend
  game_display_offset = 0
  
  Do
    game_move
    If (game_quit Or game_lost Or game_won) = false Then game_display
    utility_sleep
  Loop Until game_quit Or game_lost Or game_won
  
  game_over
End Sub

Sub game_finish
  highscore_finish
  progress_finish
  
  sound_finish
  
  alien_finish
  bomb_finish
  bonus_finish
  mapblock_finish
  player_finish
  spawner_finish
  ufo_finish
  
  For a As Integer = 1 To 128
    imagedestroy(original_graphic(a))
  Next a
End Sub

'*ORB!

Dim As Integer quit

game_start

Do
  Select Case progress_level
  Case 0, 1
    menu_main(1) = "NEW GAME"
    menu_main(2) = ""
  Case Else
    menu_main(1) = "CONTINUE GAME: LEVEL " & progress_level
    menu_main(2) = "NEW GAME"
  End Select
  
  Select Case game_menu(menu_main(), Ubound(menu_main))
  Case 1: game_play
  Case 2
    highscore_add progress_score
    progress_reset
    game_play
  Case 3: highscore_display
  Case 4: game_menu(menu_rules(), 12, true)
  Case 5: quit = true
  End Select
Loop Until quit

game_finish
System
